/*
 *
 *  makeAutomaton(rules,initialState,lastStateNumber)
 *
 *  Given a set of rules, an initial state, and the number of states to
 *  generate, create a 1D automaton.  Use the method runAutomaton() to
 *  execute the automaton and generate a series of states.
 *
 *  Warren Sack <wsack@ucsc.edu>
 *  May 2017
 *
 */

function makeAutomaton(rules,initialState,lastStateNumber) {

  ////
  //// Parameters
  ////

  // rules: rules of the rule set defined as an array of JSON structures

  // initialState: a bit vector defined as an array

  // lastStateNumber: an index for history that indicates when the 
  //                  automaton should terminate. For example, 
  //                  lastStateNumber == 100 would indicate that the 
  //                  automaton will be updated 100 times.

  ////
  //// Syntax
  ////

  //
  // isRule(obj): boolean
  //
  function isRule(obj) {
    var ks = [];
    for ( var k in obj ) { ks.push(k); }
    if ( ( ks.length === 1 ) && ( ks[0] === 'rule' ) ) {
      return(true);
    }
    else { return(false); }
  }
    
  //
  // ruleAntecendent(obj): object
  //
  function ruleAntecedent(obj) {
    return(obj["rule"]["if"]);
  }

  //
  // ruleConsequent(obj): object
  //
  function ruleConsequent(obj) {
    return(obj["rule"]["then"]);
  }

  ////
  //// Constants
  ////

  // automatonType
  var automatonType = 'elementary';
  
  ////
  //// State information
  ////

  // history: holds the successive states of the vector.  
  // history[0] corresponds to the initial state of the automaton.  
  // history[tN] corresponds to the state of the automaton at time tN.
  var history = [initialState];
  
  // vectorWidth: defines the length of the automaton.
  // We assume a "cylindrical" organization in which the left
  // edge of the vector connects to the right edge of the vector.
  var vectorWidth = history[0].length;

  // currentStateNumber: an index for history.  It is initially 0 and 
  // is updated each time a successive state of the automaton is computed.
  var currentStateNumber = 0;

  //
  // valueAtPosition(position): value
  //
  // Given a position in the vector representing the current state of the automaton,
  // returns the value in the vector of that position.  If the given position < 0, 
  // then it will return the value at vectorWidth + position.  If the given 
  // position > ( vectorWidth - 1 ), then it will return the value at 
  // position - vectorWidth.  In other words, the ends of the vector are connected 
  // together in a "cylindrical" fashion.
  //
  function valueAtPosition(position) {
    position = position % vectorWidth;
    if ( position < 0 ) { 
      return(history[currentStateNumber][vectorWidth + position]); 
    }
    if ( position > ( vectorWidth - 1 ) ) {
      return(history[currentStateNumber][position - vectorWidth]); 
    }
    else { return(history[currentStateNumber][position]); }
  }

  //
  // leftOf(position): value
  //
  function leftOf(position) {
    return(valueAtPosition(position - 1));
  }

  //
  // rightOf(position): value
  //
  function rightOf(position) {
    return(valueAtPosition(position + 1));
  }
  
  ////
  //// Database
  ////

  // database: holds the current set of rules.  Rule indices are computed from
  // the antecedents; rule consequents are stored under their indices.
  var database = { rules: {} };

  //
  // indexerElementary(left,middle,right): string
  //
  // Take numbers numbers, coerce them into strings, and concatenate them together.
  //
  function indexerElementary(left,middle,right) {
    return("" + left + middle + right);
  }

  //
  // indexerTotalistic(left,middle,right): string
  //
  // Take three numbers, compute their mean, and return it in a string.
  //
  function indexerTotalistic(left,middle,right) {
    return("" + ( ( left + middle + right ) / 3 ));
  }

  // indexer: an arity 3 function to calculate an index from the three parameters.
  // For an elementary cellular automaton, the function will append the values 
  // together and coerce it into a binary number stored in a string.  For a 
  // totalistic cellular automaton, the function will find the mean of the 
  // three input values and store it into a string.  See Stephen Wolfram, _A New
  // Kind of Science_ for definition of "elementary" and "totalistic" cellular automata.
  var indexer = {};
  if ( automatonType === 'elementary' ) { indexer = indexerElementary; }
  else { indexer = indexerTotalistic; }
  
  //
  // indexRule(rule): side-effects database["rules"]
  //
  // Accept a rule with its "if" portion containing three values.  Compute
  // a database index, "i", from those values.  Set the value of 
  // database["rules"][i] depending upon the "then" portion of the rule.
  //
  function indexRule(rule) {
    var i = indexer((1 * ruleAntecedent(rule)[0]),
		    (1 * ruleAntecedent(rule)[1]),
		    (1 * ruleAntecedent(rule)[2]));
    database["rules"][i] = ruleConsequent(rule);
  }

  //
  // fetchRule(left,middle,right): value
  //
  // Given the values of three continguous positions in the current state of the automaton,
  // compute a database index, "i", from those values.  Return the value 
  // found at database["rules"][i].
  //
  function fetchRule(left,middle,right) {
    var i = indexer((1 * left),
		    (1 * middle),
		    (1 * right));
    return(database["rules"][i]);
  }

  ////
  //// Evaluation
  ////

  //
  // initializeAutomaton()
  //
  function initializeAutomaton() {
    for ( var i = 0; i < rules.length; i++ ) {
      indexRule(rules[i]);
    }
  }

  //
  // applyRule(position,function): side-effects history[currentStateNumber + 1][position]
  //
  // Given a position in the vector representing the current state of the automaton,
  // produces a new value for that position in the next state of the automaton.
  // 
  function applyRule(position) {
    var newValueFromRule = fetchRule(leftOf(position),
				     valueAtPosition(position),
				     rightOf(position));
    history[currentStateNumber + 1][position] = newValueFromRule;
  }

  //
  // nextState(): side-effects history
  //
  // Use applyRule for position == 0 through position == vectorWidth - 1 on 
  // history[currentStateNumber].
  //
  function nextState() {
    history[history.length] = history[history.length - 1].slice(0);
    for ( var p = 0; p < vectorWidth; p++ ) {
      applyRule(p);
    }
  }

  //
  // runAutomaton()
  //
  // Run the automaton from the currentStateNumber to the lastStateNumber.
  function runAutomaton() {
    initializeAutomaton();
    while ( lastStateNumber >= currentStateNumber ) {
      nextState();
      currentStateNumber++;
    }
  }


  return({
        applyRule : applyRule,
	nextState : nextState,
	initializeAutomaton : initializeAutomaton,
	runAutomaton : runAutomaton,
        indexer : indexer,
        history : history,
	database : database,
	indexRule : indexRule,
	fetchRule : fetchRule,
        valueAtPosition : valueAtPosition,
	leftOf : leftOf,
	rightOf : rightOf,
        isRule : isRule,
	ruleAntecedent : ruleAntecedent,
	ruleConsequent : ruleConsequent	});

}






