/*
 *
 *  makeInitialState(stateID,width)
 *
 *  Create an initial state for a 1D automaton.
 *  This initial state can be programmed specially, or one can
 *  use one of the following pre-defined stateIDs:
 *  'random', 'singleBit', 'rs110ether', 'rs110gliderA',
 *  'rs110gliderA2', 'rs110gliderA3', 'rs110gliderA4', 
 *  'rs110gliderB', 'rs110AxB', 'rs110A2xB', 'rs110gliderC2',
 *  'rs110gliderEbar','rs110gliderGun'
 *
 *  Warren Sack <wsack@ucsc.edu>
 *  May 2017
 *
 */

function makeInitialState(stateID,width) {

  ////
  //// Helper functions to create various initial states
  ////

  //
  // makeVectorOfX(x,n): array
  //
  function makeVectorOfX(x,n) {
    var v = [];
    for ( var i = 0; i < n; i++ ) { v[i] = x; }
    return(v);
  }

  //
  // makeStringOfX(x,n): string
  //
  function makeStringOfX(x,n) {
    var s = "";
    for ( var i = 0; i < n; i++ ) { s = s + x; }
    return(s);
  }

  //
  // makeRandomBitVector(n): array
  //
  function makeRandomBitVector(n) {
    var v = [];
    var j = 0;
    for ( var i = 0; i < n; i++ ) {
      j = Math.random();
      if ( j >= 0.5 ) { j = 1; }
      else { j = 0; }
      v[i] = j;
    }
    return(v);
  }

  ////
  //// Variables for Rule Set 110
  ////
  var ether = "11111000100110";
  var etherLength = ether.length;
  var widthInEtherUnits = Math.floor(width / etherLength);
  var gliderA = "100110";
  var gliderA2 = "111000100110";
  var gliderA3 = "111110001110100110";
  var gliderA4 = "111110001110111000100110";
  var gliderB = "11111000";
  var gliderC2 = "11111010011100110";
  var gliderEbar = "110000010011000100110";
  var gliderGun = "11111000101100011101110011100011000100110";

  ////
  //// Set up initial state
  ////
  //// Change this to "program" the rule set, or use one
  //// of the defined states.
  ////

  var state = "";
  if ( stateID === 'random' ) {
    state = makeRandomBitVector(width);
  }
  if ( stateID === 'singleBit' ) {
    state = makeVectorOfX(0,Math.floor(width / 2));
    state[state.length] = 1;
    state = state.concat(makeVectorOfX(0,Math.floor(width / 2)));
  }
  if ( stateID === 'rs110ether' ) {
    for ( var i = 0; i < width; i += etherLength ) {
      state += ether;
    }
    state = state.split("");
  }
  if ( stateID === 'rs110gliderA' ) {
    state += ether;
    state += gliderA;
    for ( var i = 0; i < width; i += etherLength ) {
      state += ether;
    }
    state = state.split("");
  }
  if ( stateID === 'rs110gliderA2' ) {
    state += ether;
    state += gliderA2;
    for ( var i = 0; i < width; i += etherLength ) {
      state += ether;
    }
    state = state.split("");
  }
  if ( stateID === 'rs110gliderA3' ) {
    state += ether;
    state += gliderA3;
    for ( var i = 0; i < width; i += etherLength ) {
      state += ether;
    }
    state = state.split("");
  }
  if ( stateID === 'rs110gliderA4' ) {
    state += ether;
    state += gliderA4;
    for ( var i = 0; i < width; i += etherLength ) {
      state += ether;
    }
    state = state.split("");
  }
  if ( stateID === 'rs110gliderB' ) {
    for ( var i = 0; i < ( width - ( 2 * etherLength ) ); i += etherLength ) {
      state += ether;
    }
    state += gliderB;
    state += ether;
    state = state.split("");
  }
  if ( stateID === 'rs110AxB' ) {
    state += ether;
    state += gliderA;
    for ( var i = 0; i < ( width - ( 4 * etherLength ) ); i += etherLength ) {
      state += ether;
    }
    state += gliderB;
    state += ether;
    state = state.split("");
  }
  if ( stateID === 'rs110A2xB' ) {
    state += ether;
    state += gliderA2;
    for ( var i = 0; i < ( width - ( 4 * etherLength ) ); i += etherLength ) {
      state += ether;
    }
    state += gliderB;
    state += ether;
    state = state.split("");
  }
  if ( stateID === 'rs110gliderC2' ) {
    state += ether;
    state += gliderC2;
    for ( var i = 0; i < width; i += etherLength ) {
      state += ether;
    }
    state = state.split("");
  }
  if ( stateID === 'rs110gliderEbar' ) {
    for ( var i = 0; i < ( width - ( 2 * etherLength ) ); i += etherLength ) {
      state += ether;
    }
    state += gliderEbar;
    state += ether;
    state = state.split("");
  }
  if ( stateID === 'rs110gliderGun' ) {
    for ( var i = 0; i < Math.floor(widthInEtherUnits / 2); i += 1 ) {
      state += ether;
    }
    state += gliderGun;
    for ( var i = 0; i < Math.floor(widthInEtherUnits / 2); i += 1 ) {
      state += ether;
    }
    state = state.split("");
  }

  return(state);
}
