/*
 *  makeLifeRuleSet()
 *  
 *  Computes a set of 512 if-then rules that 
 *  implement Conway's "Game of Life."  These
 *  rules can be described more compactly like this:
 *  
 *  Rules of Life
 *  birth:
 *  0 --> 1 iff 3 neighbors are 1
 *  death by overcrowding:
 *  1 --> 0 if 4 or 4+ of neighbors are 1
 *  death by exposure:
 *  1 --> 0 if zero or 1 neighbors are 1
 *  survival:
 *  1 --> 1 if 2 or 3 neighbors are 1
 *  
 *  Warren Sack <wsack@ucsc.edu>
 *  May 2017
 *
 */

function makeLifeRuleSet() {

  // leftPad(c,s,l): string
  //
  // Pad the left of a string with multiple copies of c
  // until the string, s, is of length l.
  //
  function leftPad(c,s,l) {
    if ( s.length >= l ) { return(s); }
    else { return(leftPad(c,c+s,l)); }
  }

  //
  // leftStrip(c,s): string
  //
  // Strip off all copies of c from the left end of string, s.
  //
  function leftStrip(c,s) {
    if ( s[0] !== c ) { return(s); }
    else { return(leftStrip(c,s.substr(1,s.length-1))); }
  }

  //
  // decimalToBinary(n): binary number as string
  //
  function decimalToBinary(n) {
    var binaryString = '';
    if ( n === 0 ) { return('0'); }
    while ( n > 0 ) {
      binaryString = ( n % 2 ) + binaryString;
      n = n >> 1;
    }
    return(binaryString);
  }

  //
  // makeRuleSet(): array of objects
  //
  function makeRuleSet() {
    var rs = [];
    var rb = "";
    var consequent = "";
    var nextRule = {};
    for (var r = 511; r >= 0; r-- ) {
      consequent = "";
      rb = decimalToBinary(r);
      // leftPad rb so that it contains 9 chars
      rb = leftPad("0",rb,9);
      var sumOfNeighbors = 
	( rb.charAt(0) * 1 ) 
	+ ( rb.charAt(1) * 1 ) 
	+ ( rb.charAt(2) * 1 ) 
	+ ( rb.charAt(3) * 1 ) 
	+ ( rb.charAt(5) * 1 ) 
	+ ( rb.charAt(6) * 1 ) 
	+ ( rb.charAt(7) * 1 ) 
	+ ( rb.charAt(8) * 1 );
      // birth
      if ( rb.charAt(4) === "0" ) {
	if ( sumOfNeighbors === 3 ) { consequent = "1"; }
	else { consequent = "0"; }
      }
      else {   
	if ( rb.charAt(4) === "1" ) {
	  // death by overcrowding
	  if ( sumOfNeighbors >= 4 ) { consequent = "0"; }
	  else { 
	    // death by exposure
	    if ( sumOfNeighbors <= 1 ) { consequent = "0"; }
	    else {
	      // survival
	      if ( ( sumOfNeighbors === 2 )
		   || ( sumOfNeighbors === 3 ) ) { 
		consequent = "1"; 
	      }
	    }
	  }
	}
      }
      rb = rb.split("");
      nextRule = {"rule": {"if": rb, "then": consequent}};
      rs.push(nextRule);
    }
    return(rs);
  }
  
  return({ makeRuleSet : makeRuleSet});
}
