/*
 *  makeAutomaton(rules,initialState)
 *
 *  Make a 2D automaton given a rule set and an initialState.
 *
 * Warren Sack <wsack@ucsc.edu>
 * May 2017
 *
 */

function makeAutomaton(rules,initialState) {

  ////
  //// Syntax
  ////

  //
  // isRule(obj): boolean
  //
  function isRule(obj) {
    var ks = [];
    for ( var k in obj ) { ks.push(k); }
    if ( ( ks.length === 1 ) && ( ks[0] === 'rule' ) ) {
      return(true);
    }
    else { return(false); }
  }
    
  //
  // ruleAntecendent(obj): object
  //
  function ruleAntecedent(obj) {
    return(obj["rule"]["if"]);
  }

  //
  // ruleConsequent(obj): object
  //
  function ruleConsequent(obj) {
    return(obj["rule"]["then"]);
  }

  ////
  //// Constants
  ////

  // automatonType
  var automatonType = 'elementary';

  
  ////
  //// State information
  ////

  //
  // copy2DArray(a): array
  //
  function copy2DArray(a) {
    var c = [];
    for ( var i = 0; i < a.length; i++ ) {
      for ( var j = 0; j < a[i].length; j++ ) {
	c[i] = a[i].slice(0);
      }
    }
    return(c);
  }
	

  //
  // getCurrentStateNumber(): integer
  //
  function getCurrentStateNumber() {
    return(currentStateNumber);
  }

  // initialState contains the first state of the automaton
  // currentState is the current state of the automaton, written as a 2D array
  var currentState = copy2DArray(initialState);

  //
  // getCurrentState(): array
  //
  function getCurrentState() {
    return(currentState);
  }

  //
  // setCurrentState(state): side-effects currentState
  //
  function setCurrentState(state) {
    currentState = copy2DArray(state);
    return(currentState);
  }

  // nextState is used to hold the values being computed for the next state of
  // the automaton.
  var nextState = copy2DArray(initialState);
  
  // arrayWidth defines the width of the automaton.
  // We assume a "cylindrical" organization in which the left
  // edge of the array connects to the right edge of the array.
  var arrayWidth = initialState[0].length;

  // arrayHeight defines the height of the automaton.
  // We assume a "cylindrical" organization in which the top
  // edge of the array connects to the bottom edge of the array.
  var arrayHeight = initialState.length;

  // currentStateNumber is initially 0 and is updated each time a
  // successive state of the automaton is computed.
  var currentStateNumber = 0;

  //
  // getValueAtPosition(x,y): value
  //
  // Given a position in the array representing the current state of the automaton,
  // returns the value in the array of that position.  If the x of the given position < 0, 
  // then it will return the value at arrayWidth + x.  If the given 
  // x > ( arrayWidth - 1 ), then it will return the value at x - arrayWidth.  
  // In other words, the ends of the array are connected together in a "cylindrical" fashion.
  // An analogous arrangement pertains to the top and the bottom of the array.
  //
  function getValueAtPosition(x,y) {
    x = x % arrayWidth;
    if ( x < 0 ) { x = arrayWidth + x; }
    if ( x > ( arrayWidth - 1 ) ) { x = x - arrayWidth; }
    y = y % arrayHeight;
    if ( y < 0 ) { y = arrayHeight + y; }
    if ( y > ( arrayHeight - 1 ) ) { y = y - arrayHeight; }
    return(currentState[x][y]);
  }

  //
  // setValueAtPosition(x,y,v): side-effects currentState[x][y]
  //
  function setValueAtPosition(x,y,v) {
    x = x % arrayWidth;
    if ( x < 0 ) { x = arrayWidth + x; }
    if ( x > ( arrayWidth - 1 ) ) { x = x - arrayWidth; }
    y = y % arrayHeight;
    if ( y < 0 ) { y = arrayHeight + y; }
    if ( y > ( arrayHeight - 1 ) ) { y = y - arrayHeight; }
    currentState[x][y] = v;
    return(currentState[x][y]);
  }

  //
  // leftOf(x,y): value
  //
  function leftOf(x,y) {
    return(getValueAtPosition(x - 1,y));
  }

  //
  // rightOf(x,y): value
  //
  function rightOf(x,y) {
    return(getValueAtPosition(x + 1,y));
  }

  //
  // above(x,y): value
  //
  function above(x,y) {
    return(getValueAtPosition(x,y - 1));
  }

  //
  // below(x,y): value
  //
  function below(x,y) {
    return(getValueAtPosition(x,y + 1));
  }
  
  //
  // aboveLeftOf(x,y): value
  //
  function aboveLeftOf(x,y) {
    return(getValueAtPosition(x - 1,y - 1));
  }

  //
  // aboveRightOf(x,y): value
  //
  function aboveRightOf(x,y) {
    return(getValueAtPosition(x + 1,y - 1));
  }

  //
  // belowLeftOf(x,y): value
  //
  function belowLeftOf(x,y) {
    return(getValueAtPosition(x - 1,y + 1));
  }

  //
  // belowRightOf(x,y): value
  //
  function belowRightOf(x,y) {
    return(getValueAtPosition(x + 1,y + 1));
  }

  ////
  //// Database
  ////

  // database holds the current set of rules.  Rule indices are computed from
  // the antecedents; rule consequents are stored under their indices.
  var database = { rules: {} };

  //
  // indexRule(rule): side-effects database["rules"]
  //
  // Accept a rule with its "if" portion containing three values.  Compute
  // a database index, "i", from those values.  Set the value of 
  // database["rules"][i] depending upon the "then" portion of the rule.
  //
  function indexRule(rule) {
    var i = ruleAntecedent(rule).toString();
    i = i.replace(/\,/,"");
    database["rules"][i] = ruleConsequent(rule);
  }

  //
  // fetchRule(ps): value
  //
  // Given the values of nine ( 3 x 3 ) continguous positions in the current state 
  // of the automaton, compute a database index, "i", from those values.  Return the value 
  // found at database["rules"][i].
  //
  function fetchRule(ps) {
    var i = ps.toString();
    i = i.replace(/\,/,"");
    return(database["rules"][i]);
  }

  ////
  //// Evaluation
  ////

  //
  // initializeAutomaton()
  //
  function initializeAutomaton() {
    for ( var i = 0; i < rules.length; i++ ) {
      indexRule(rules[i]);
    }
  }

  //
  // applyRule(x,y): side-effects nextState[x,y]
  //
  // Given a position in the array representing the current state of the automaton,
  // produces a new value for that position in the next state of the automaton.
  // 
  function applyRule(x,y) {
    var subArray = [aboveLeftOf(x,y),above(x,y),aboveRightOf(x,y),
		    leftOf(x,y),getValueAtPosition(x,y),rightOf(x,y),
		    belowLeftOf(x,y),below(x,y),belowRightOf(x,y)];
    var newValueFromRule = fetchRule(subArray);
    nextState[x][y] = newValueFromRule;
  }

  //
  // computeNextState(): side-effects nextState
  //
  // Use applyRule for each element of currentState.
  //
  function computeNextState() {
    var x = 0;
    var y = 0;
    // rewrite the nextState applying rules to currentState
    for ( x = 0; x < arrayWidth; x++ ) {
      for ( y = 0; y < arrayHeight; y++ ) {
	applyRule(x,y);
      }
    }
    // Assign values of nextState to currentState
    for ( x = 0; x < arrayWidth; x++ ) {
      for ( y = 0; y < arrayHeight; y++ ) {
	currentState[x][y] = nextState[x][y];
      }
    }
    currentStateNumber++;    
  }

  //
  // runAutomaton()
  //
  // Run the automaton.
  //
  function runAutomaton() {
    initializeAutomaton();
    computeNextState();
  }


  return({
        getCurrentStateNumber : getCurrentStateNumber,
        getCurrentState : getCurrentState,
        setCurrentState : setCurrentState,
	arrayWidth : arrayWidth,
	arrayHeight : arrayHeight,
        applyRule : applyRule,
	computeNextState : computeNextState,
	initializeAutomaton : initializeAutomaton,
	runAutomaton : runAutomaton,
	database : database,
	indexRule : indexRule,
	fetchRule : fetchRule,
        getValueAtPosition : getValueAtPosition,
        setValueAtPosition : setValueAtPosition,
	leftOf : leftOf,
	rightOf : rightOf,
        isRule : isRule,
	ruleAntecedent : ruleAntecedent,
	ruleConsequent : ruleConsequent	});

}






