/*
 *  makeDisplay(automaton,canvasName)
 *
 *  Given an 2D automaton and a canvasName, update and display
 *  the current state of the automaton.
 *
 *  Pressing the <space> bar will start or stop the execution of the
 *  automaton.
 *
 *  Pressing the <p> key will save the current state of the display 
 *  into a PNG file.
 *
 *  Clicking the mouse down on a square will recolor it; if it is
 *  white, it will be recolored black; if it is black it will be 
 *  recolored white.  The current state of the automaton is updated
 *  to reflect the change in color (where black is "1" and white is "0").
 *
 *  Warren Sack <wsack@ucsc.edu>
 *  May 2017
 *
 */
function makeDisplay(automaton,canvasName) {

  window.resizeTo(screen.width,screen.height);
  var width = automaton.arrayWidth;
  var height = automaton.arrayHeight;
  var lineWidth = 1;
  var margin = screen.availTop;
  var squareWidth = ( screen.width - ( 2 * margin ) ) / width;
  var squareHeight = ( screen.height - ( 2 * margin ) ) / ( 1.3 * height );
  if ( squareWidth > squareHeight ) { squareWidth = squareHeight; }
  var canvas = document.getElementById(canvasName);
  canvas.width = ( squareWidth * width ) + ( 2 * margin );
  canvas.height = ( squareWidth * height ) + ( 2 * margin );
  window.addEventListener('mousedown',recolorSquare,false);
  window.addEventListener('keydown',startStop,false);
  window.addEventListener('keydown',snapShot,false);
  var context = canvas.getContext('2d');
  var running = false;
  var timeBetweenFrames = 100;
  var intervalID = setInterval(updateAndDisplay,timeBetweenFrames);
  if (! running) { clearInterval(intervalID); }

  //
  // updateAndDisplay(): side-effects the DOM and the current
  //                     state of the automaton
  //
  function updateAndDisplay() {
    automaton.computeNextState(); 
    drawMatrix();
   }

  //
  // startStop(event): sets or clears the interval
  //
  function startStop(event) {
    if (event.keyCode == 32) {
      if ( running ) { 
	clearInterval(intervalID); 
	running = false;
      }
      else {
	intervalID = setInterval(updateAndDisplay,timeBetweenFrames);
	running = true;
      }
    }
  }

  //
  // snapShot(event): generates a PNG file
  //
  function snapShot(event) {
    if (event.keyCode == 80) {
       image = new Image();
       image = canvas.toDataURL("image/png");
       window.open(image,"snapShot");
       window.focus();
    }
  }

  //
  // recolorSquare(event): side-effects the DOM and the current
  //                       state of the automaton
  //
  function recolorSquare(event) {
    var xPos = event.clientX;
    var yPos = event.clientY;
    // translate this to an [column][row] position for the matrix
    var column = -1;
    var row = -1;
    var x = margin;
    for ( var c = 0; c < width; c++ ) {
      x = x + squareWidth;
      if ( x > xPos ) { column = c; break; }
    }
    var y = margin;
    for ( var r = 0; r < height; r++ ) {
      y = y + squareWidth;
      if ( y > yPos ) { row = r; break; }
    }
    // Set the square to "1" if it is "0" and vice versa.
    // And, recolor the square appropriately (black to white;
    // or, white to black).
    if ( ( row !== -1 ) && ( column !== -1 ) ) {
      x = x - squareWidth;
      y = y - squareWidth;
      context.strokeRect(x,y,squareWidth,squareWidth);
      if ( automaton.getValueAtPosition(column,row) === "1" ) {
	automaton.setValueAtPosition(column,row,"0");
	context.fillStyle = 'white';
      }
      else { 
	automaton.setValueAtPosition(column,row,"1"); 
	context.fillStyle = 'black';
      }
      context.fillRect((x + lineWidth),
		       (y + lineWidth),
		       (squareWidth - (2 * lineWidth)),
		       (squareWidth - (2 * lineWidth)));
    }
  }

  //
  // drawMatrix(): side-effects the DOM
  //
  function drawMatrix() {
    var x = margin;
    var y = margin;
    context.lineWidth = lineWidth;
    context.strokeStyle = 'gray';
    for ( var row = 0; row < height; row++ ) {
      for ( var column = 0; column < width; column++ ) {
       	context.strokeRect(x,y,squareWidth,squareWidth);
	if ( automaton.getValueAtPosition(column,row) === "1" ) {
	  context.fillStyle = 'black';
	}
	else { context.fillStyle = 'white'; }
	context.fillRect((x + lineWidth),
			 (y + lineWidth),
			 (squareWidth - (2 * lineWidth)),
			 (squareWidth - (2 * lineWidth)));
	x = x + squareWidth;
      }
      x = margin;
      y = y + squareWidth;
    }
  }

  return({
        drawMatrix : drawMatrix,
	updateAndDisplay : updateAndDisplay});
 
}
