/*
 * makeCircuitSimulator
 *
 * Warren Sack <wsack@ucsc.edu>
 * May 2017
 *
 * Simulate components of boolean circuits.
 * This is a translation from Scheme into JavaScript
 * of the circuit simulator in section 3.3.4 of Sussman's
 * and Abelson's SICP book, _The Structure and Interpretation
 * of Computer Programs_.
 *
 */

function makeCircuitSimulator() {

  function makeQueue(as) {
    var actions = as;
    function isEmpty() { return(actions.length === 0); }
    function addAction(a) { actions.push(a); }
    function getFirstAction() { return(actions[0]); }
    function removeFirstAction() { actions.shift(); }
    return({isEmpty: isEmpty,
 	    addAction: addAction,
            getFirstAction: getFirstAction,
	    removeFirstAction: removeFirstAction});
  }

  function makeAgenda() {
    var currentTime = 0;
    var queues = {};
    function getCurrentTime() { return(currentTime); }
    function setCurrentTime(t) { currentTime = t; }
    function isEmpty() {
      var ks = [];
      for ( var k in queues ) { ks.push(k); }
      return( ks.length === 0 );
    }
    function getFirstTime() {
      var ks = [];
      for ( var k in queues ) { ks.push(k); }
      return(ks.sort()[0]);
    }
    function getFirstAction() {
      var t = getFirstTime();
      var a = queues[t].getFirstAction();
      setCurrentTime(t);
      return(a);
    }
    function removeFirstAction() {
      var t = getFirstTime();
      queues[t].removeFirstAction();
      if ( queues[t].isEmpty() ) { delete queues[t]; }
    }
    function addAction(a,t) {
      var q = queues[t];
      if ( q === undefined ) { q = makeQueue([]); }
      q.addAction(a);
      queues[t] = q;
    }
    function addActionWithDelay(a,d) {
      var t = getCurrentTime();
      var after = (t - 0) +  (d - 0);
      addAction(a,after);
    }
    function propagate() {
      if (! isEmpty() ) {
	var a = getFirstAction();
	a.call();
	removeFirstAction();
	propagate();
      }
    }
    return({getCurrentTime: getCurrentTime,
 	    setCurrentTime: setCurrentTime,
	    isEmpty: isEmpty,
	    getFirstAction: getFirstAction,
	    removeFirstAction: removeFirstAction,
	    addAction: addAction,
	    addActionWithDelay: addActionWithDelay,
	    propagate: propagate});
  }

  var agenda = makeAgenda();
  
  function getAgenda() { return(agenda); }

  function makeWire() {
    var value = 0;
    var actions = [];
    function getValue() { return(value); }
    function setValue(v) {
      if ( v !== value ) {
	value = v;
	for (var i = 0; i < actions.length; i++) {
	  actions[i].call();
	}
      }
    }
    function addAction(a) { 
      actions.push(a); 
      a.call();
    }
    function probe(name) {
      function report() {
	console.log("value for " + 
		    name + " @ time " + 
		    agenda.getCurrentTime() + 
		    " is " + getValue());
      }
      addAction(report);
    }
    return({getValue: getValue,
	    setValue: setValue,
	    addAction: addAction,
	    probe: probe});
  }

  function logicalNot(v) {
    if ( v === 1 ) { return(0); }
    else { return(1); }
  }

  function logicalAnd(v1,v2) {
    return(v1 && v2);
  }

  function logicalOr(v1,v2) {
    return(v1 || v2);
  }

  function inverter(i,o) {
    var delay = 3;
    function invertInput() {
      var v = logicalNot(i.getValue());
      var a = function() { o.setValue(v); };
      agenda.addActionWithDelay(a,delay);
    }
    i.addAction(invertInput);
  }

  function relay(i,o,delay) {
    function relayInput() {
      var a = function() { o.setValue(i.getValue()); };
      agenda.addActionWithDelay(a,delay);
    }
    i.addAction(relayInput);
  }

  function andGate(i1,i2,o) {
    var delay = 3;
    function andAction() {
      var v = logicalAnd(i1.getValue(),i2.getValue());
      var a = function() { o.setValue(v); };
      agenda.addActionWithDelay(a,delay);
    }
    i1.addAction(andAction);
    i2.addAction(andAction);
  }

  function orGate(i1,i2,o) {
    var delay = 5;
    function orAction() {
      var v = logicalOr(i1.getValue(),i2.getValue());
      var a = function() { o.setValue(v); };
      agenda.addActionWithDelay(a,delay);
    }
    i1.addAction(orAction);
    i2.addAction(orAction);
  }

  return({getAgenda: getAgenda,
	  makeWire: makeWire,
  	  inverter: inverter,
	  relay: relay,
	  andGate: andGate,
	  orGate: orGate});
}



